/********************************************************************\

  Name:         caen170a.c
  Created by:   Stefan Ritt

  Contents:     CAEN 170a High Voltage Device Driver


  Revision history
  -------------------------------------------------------------------
  date         by    modification
  ---------    ---   ------------------------------------------------
  26-MAR-97    SR    created
  15-Apr-97    PAA   from lrs4032

\********************************************************************/

#include "midas.h"
#include "msystem.h"
#if __NORPC__
 #include "cam8210.h"
#endif

#define FRAME_TO_CHANNEL 0
#define CHANNEL_TO_FRAME 1

#define CAEN170A_STR "\
BCN_Mainframe = STRING : [32] 0 2 19 11 21\n\
"
/*---- globals -----------------------------------------------------*/
void string_parse(char *line, int *ntoken, char *p,char *token[]);

/* Specific to the CAEN170A HV control structure */
static struct {
  DWORD  ext_hv;
  INT    mainframe[16];
  DWORD  branch;
  DWORD  crate;
  DWORD  station;
} caen;

/*---- device driver routines ------------------------------------------------*/

INT caen170a_init(HNDLE hKeyRoot, INT channels)
{
  char chain_address[256];
  int   status, size, i, n_cards, q;
  char  str[256];
  HNDLE hDB;
  
  cm_get_experiment_database(&hDB, NULL);
  
  /* Get Mainframe address */
  {
    INT ntoken, i;
    char * token[16], *p;
    
    status = db_create_record(hDB, hKeyRoot, "", CAEN170A_STR);
    if (status != DB_SUCCESS)
      {
	cm_msg(MERROR,"caen170a_init","Cannot create BCN_Mainframe\n");
	return FE_ERR_ODB;
      }
    size = sizeof(chain_address);
    status = db_get_value(hDB, hKeyRoot, "BCN_Mainframe", chain_address, &size, TID_STRING);
    if (status != DB_SUCCESS)
      {
	cm_msg(MINFO, "CAEN170a_init", "Cannot find 'CAEN170a_Mainframe BCN_M' entry in /hv/CAEN170A database");
	return FE_ERR_ODB;
      }
    p = (char *) calloc(256,sizeof(char));
    string_parse(chain_address, &ntoken,  p, token);
    caen.branch = atoi(token[0]);
    caen.crate  = atoi(token[1]);
    caen.station= atoi(token[2]);
    printf("CAEN Mainframe BCN_M %s - %ld %ld %ld - \n",chain_address, caen.branch,caen.crate,caen.station);
    if (ntoken < 4)
      {
	printf(" Mainframe not defined:%s\n",chain_address);
	cm_msg(MINFO,  "CAEN170a_init","Mainframe not defined in database");
	return FE_ERR_ODB;
      }    
    for (i=3;i<ntoken;i++)
      {
	caen.mainframe[i-3] = atoi(token[i]);
	printf("Valid CAEN Mainframe address:%i\n",caen.mainframe[i-3]);
      }
    free (p);
  }

  /* read LRS1440 settings */
  /* initialize RS232 port */
  /* check if module is living  */
  /* check if HV enabled */
  /* read rest of status */
  /* turn on HV main switch */
  
  /* save ext for HV slot */
#if __NORPC__
  CAM_BC (&caen.ext_hv,caen.branch,caen.crate);

  /* Init CAEN170A */
  CAM_COM(caen.branch,caen.crate,caen.station,0,24);  /* Disable LAM */
  CAM_CMD(&caen.ext_hv,caen.station,0,9);
#else
  caopen("vwchaos",&status);
  cdreg(&caen.ext_hv,caen.branch,caen.crate,caen.station,0);
  printf("Status:%d -E%xB%dC%dN%d\n",status,caen.ext_hv,caen.branch,caen.crate,caen.station);
  cfsa(24,caen.ext_hv,&status,&q);
  cfsa(9,caen.ext_hv,&status,&q);
#endif

  return FE_SUCCESS;
}

/*----------------------------------------------------------------------------*/

INT caen170a_exit()
{
  printf("in caen170a_exit \n");
  return FE_SUCCESS;
}

/*----------------------------------------------------------------------------*/
INT caen170a_assign_channel (INT direction, INT *channel, DWORD *frame, DWORD *chanframe)
{
  DWORD   ch;

  ch = *channel;

  switch (direction)
    {
    case FRAME_TO_CHANNEL:
      printf("caen170a_assign_channel: FRAME_TO_CHANNEL not implemented\n");
      break;
     case CHANNEL_TO_FRAME:
      *frame  = caen.mainframe[(ch / 64 )];
      *chanframe = (ch % 64);
#ifdef DEBUG
      printf("%i ->",ch);
      printf(" %i %i ",*frame,*chanframe);
#endif
      break;
    }
  return FE_SUCCESS;
}

/*----------------------------------------------------------------------------*/
INT caen170a_set(INT channel, float fvalue)
{
  DWORD cmd, frame, fch, data;
  DWORD nlam, q, x;
  INT   status, value;
  WORD  svalue;

  value = (INT)abs(fvalue);

  /* convert channel to frame and channel within frame */
  if (caen170a_assign_channel(CHANNEL_TO_FRAME,&channel,&frame,&fch) < 0)
    return FE_ERR_ODB;
  
  /* Clear buffer */
#if __NORPC__
  CAM_CMD(&caen.ext_hv,caen.station,0,9);
#else
  cdreg(&caen.ext_hv,caen.branch,caen.crate,caen.station,0);
  cfsa(9,caen.ext_hv,&status,&q);
#endif

  /* HAS to be there! */
  ss_sleep(400);

  /* host */
#if __NORPC__
  CAM24_OQX(&caen.ext_hv,caen.station,0,16,1,&q,&x);
#else
  cdreg(&caen.ext_hv,caen.branch,caen.crate,caen.station,0);
  status = 1;
  cfsa(16,caen.ext_hv,&status,&q);
#endif

  /* crate */
#if __NORPC__
  CAM24_OQX(&caen.ext_hv,caen.station,0,16,frame,&q,&x);
#else
  cdreg(&caen.ext_hv,caen.branch,caen.crate,caen.station,0);
  cfsa(16,caen.ext_hv,&frame,&q);
#endif

  /* operation 0x05.. Set relay OFF , .. channel#*/
  /* operation 0x04.. Set relay ON , .. channel#*/
  /* operation 0x03.. write Vset, .. channel#*/
  cmd = 0x0300 | (fch & 0xff);
#if __NORPC__
  CAM24_OQX(&caen.ext_hv,caen.station,0,16,cmd,&q,&x);
#else
  cdreg(&caen.ext_hv,caen.branch,caen.crate,caen.station,0);
  cfsa(16,caen.ext_hv,&cmd,&q);
#endif

  /* set HV */
#if __NORPC__
  CAM24_OQX(&caen.ext_hv,caen.station,0,16,value,&q,&x);
#else
  cdreg(&caen.ext_hv,caen.branch,caen.crate,caen.station,0);
  cfsa(16,caen.ext_hv,&value,&q);
#endif

  /* ship op request */
#if __NORPC__
  CAM24_OQX(&caen.ext_hv,caen.station,0,17,1,&q,&x);
#else
  cdreg(&caen.ext_hv,caen.branch,caen.crate,caen.station,0);
  status = 1;
  cfsa(17,caen.ext_hv,&status,&q);
#endif

  /* HAS to be there! */
  ss_sleep(300);
  nlam = 0;
  do
    {
      ss_sleep(100);
#if __NORPC__
      CAM16_IQX(&caen.ext_hv,caen.station,0,0,&svalue,&q,&x);
#else
      cdreg(&caen.ext_hv,caen.branch,caen.crate,caen.station,0);
      cfsa(0,caen.ext_hv,&svalue,&q);
#endif
      nlam++;
    } while ((nlam < 10) && (q != 1));

  /* decode value */
  /* error returned :
     0xff00 : busy.
     0xff01 : unknown code or erro in the request.
     0xff02 : no channel present.
     0xff03 : Vset > HVin.
     0xff04 : HVin > 1500 ON command not enabled.
     0xff05 : Config number > 9.
     0xfffd : No data transmitted.
     0xfffe : The master ID is incorrect.
     0xffff : The address slave does not exist.
  */
#if __NORPC__
  CAM16_IQX(&caen.ext_hv,caen.station,0,0,&svalue,&q,&x);
#else
  cdreg(&caen.ext_hv,caen.branch,caen.crate,caen.station,0);
  cfsa(0,caen.ext_hv,&svalue,&q);
#endif

#ifdef DEBUG
  printf(" - value :0x%x - value:%i - qx:%i%i : nlam:%i\n",svalue, svalue,q,x,nlam); 
#endif

  return FE_SUCCESS;
}

/*----------------------------------------------------------------------------*/

INT caen170a_set_all(INT channels, float value)
{
  INT   i;
  
  for (i=0 ; i<channels ; i++)
    {
      caen170a_set(i, value);
      cm_yield(0); /* check for watchdog requests */
    }

  return FE_SUCCESS;
}

/*----------------------------------------------------------------------------*/

INT caen170a_get(INT channel, float *pvalue)
{
  DWORD cmd, frame, fch;
  DWORD nlam, q, x;
  DWORD value;

  /* convert channel to frame and channel within frame */
  if (caen170a_assign_channel(CHANNEL_TO_FRAME,&channel,&frame,&fch) < 0)
    return FE_ERR_ODB;

  /* clear LAM */
#if __NORPC__
  CAM_CMD(&caen.ext_hv,caen.station,0,9);
#else
  cdreg(&caen.ext_hv,caen.branch,caen.crate,caen.station,0);
  cfsa(9,caen.ext_hv,&cmd,&q);
#endif

  /* HAS to be there! */
  ss_sleep(400);

  /* host = 1 */
#if __NORPC__
  CAM24_OQX(&caen.ext_hv,caen.station,0,16,1,&q,&x);
#else
  cdreg(&caen.ext_hv,caen.branch,caen.crate,caen.station,0);
  cmd = 1;
  cfsa(16,caen.ext_hv,&cmd,&q);
#endif

  /* mainframe CAEN address */
#if __NORPC__
  CAM24_OQX(&caen.ext_hv,caen.station,0,16,frame,&q,&x);
#else
  cdreg(&caen.ext_hv,caen.branch,caen.crate,caen.station,0);
  cfsa(16,caen.ext_hv,&frame,&q);
#endif

  /* operation 0x0000 Read Module name */
  /* operation 0x02.. Read Vset , .. channel# */
  /* operation 0x01.. Read Vmon , .. channel# */
  cmd = 0x0100 | (fch & 0xff);
#if __NORPC__
  CAM24_OQX(&caen.ext_hv,caen.station,0,16,cmd,&q,&x);
#else
  cdreg(&caen.ext_hv,caen.branch,caen.crate,caen.station,0);
  cfsa(16,caen.ext_hv,&cmd,&q);
#endif

  /* ship op request */
#if __NORPC__
  CAM24_OQX(&caen.ext_hv,caen.station,0,17,1,&q,&x);
#else
  cdreg(&caen.ext_hv,caen.branch,caen.crate,caen.station,0);
  cmd = 1;
  cfsa(17,caen.ext_hv,&cmd,&q);
#endif

  /* HAS to be there! */
  ss_sleep(400);

  nlam = 0;
  do 
    {
      ss_sleep(100);
#if __NORPC__
      CAM24_IQX(&caen.ext_hv,caen.station,0,0,&value,&q,&x);
#else
      cdreg(&caen.ext_hv,caen.branch,caen.crate,caen.station,0);
      cfsa(0,caen.ext_hv,&value,&q);
#endif
      nlam++;
    } while ((nlam < 10) && (q != 1));

  /* read data & decode value remove relay status (bit 15) if below error status */
#if __NORPC__
  CAM24_IQX(&caen.ext_hv,caen.station,0,0,&value,&q,&x);
#else
  cdreg(&caen.ext_hv,caen.branch,caen.crate,caen.station,0);
  cfsa(0,caen.ext_hv,&value,&q);
#endif
  /* error returned :
     0xff00 : busy.
     0xff01 : unknown code or erro in the request.
     0xff02 : no channel present.
     0xff03 : Vset > HVin.
     0xff04 : HVin > 1500 ON command not enabled.
     0xff05 : Config number > 9.
     0xfffd : No data transmitted.
     0xfffe : The master ID is incorrect.
     0xffff : The address slave does not exist.
  */
  if (value < 0xff00) 
    *pvalue = (float)(value & 0x7fff);

#ifdef DEBUG
  printf(" - value :0x%x - value:%f - qx:%i%i : nlam:%i\n",value, *pvalue,q,x,nlam); 
#endif
  
  return FE_SUCCESS;
}

/*----------------------------------------------------------------------------*/

INT caen170a_get_all(INT channels, float *array)
{
  float   *pvalue;
  int      i;
  

 /* read ALL this equip channels */
  for (i=0 ; i <= channels ; i++)
    {
      caen170a_get(i, pvalue);
      array[i] = *pvalue;
      cm_yield(0); /* check for watchdog requests */
    }
  
  return FE_SUCCESS;
}

/*----------------------------------------------------------------------------*/

/*---- device driver entry point -----------------------------------*/

INT caen170a(INT cmd, ...)
{
  va_list argptr;
  HNDLE   hKey;
  INT     channel, status;
  float   value, *pvalue;
  
  va_start(argptr, cmd);
  status = FE_SUCCESS;
  
  switch (cmd)
    {
    case CMD_INIT:
      hKey = va_arg(argptr, HNDLE);
      channel = va_arg(argptr, INT);
      status = caen170a_init(hKey, channel);
      break;
      
    case CMD_EXIT:
      status = caen170a_exit();
      break;
      
    case CMD_SET:
     channel = va_arg(argptr, INT);
      value   = (float) va_arg(argptr, double);
#ifdef DEBUG
      printf("in caen170a cmd set ->:%i -chan:%i -Val:%f ",cmd,channel,value);
#endif
      status = caen170a_set(channel, value);
     break;
      
    case CMD_SET_ALL:
      channel = va_arg(argptr, INT);
      value   = (float) va_arg(argptr, double);
#ifdef DEBUG
      printf("in caen170a cmd set all:%i -chan:%i -Val:%f ",cmd,channel,value);
#endif
      status = caen170a_set_all(channel, value);
      break;
      
    case CMD_GET:
      channel = va_arg(argptr, INT);
      pvalue  = va_arg(argptr, float*);
      status = caen170a_get(channel, pvalue);
#ifdef DEBUG
      printf("in caen170a cmd get <-:%i -Chan:%i -Val:%f ",cmd,channel,*pvalue);
#endif
      break;
      
    case CMD_GET_ALL:
      channel = va_arg(argptr, INT);
      pvalue  = va_arg(argptr, float*);
      status = caen170a_get_all(channel, pvalue);
#ifdef DEBUG
      printf("in caen170a cmd get all:%i -Chan:%i -Val:%f ",cmd,channel,*pvalue);
#endif
      break;
      
      /* current measurements not supported in CAEN170A */
      
    case CMD_GET_CURRENT:
      break;
      
    case CMD_GET_CURRENT_ALL:
      break;
      
    case CMD_SET_CURRENT_LIMIT:
      break;
      
    default:
      cm_msg(MERROR, "caen170a device driver", "Received unknown command %d", cmd);
      status = FE_ERR_DRIVER;
      break;
    }
  
  va_end(argptr);
  
  return status;
}

/*------------------------------------------------------------------*/
