/********************************************************************\

  Name:         mhyt1331.h
  Created by:   Stefan Ritt

  Cotents:      Declarations for CAMAC functions for HYTEC 1331
                controller under MS Visual C++.

  Revision history
  ------------------------------------------------------------------
  date        by    modification
  --------    ---   ------------------------------------------------
  22-JAN-95   SR    created

\********************************************************************/

/*------------------------------------------------------------------*/

/*
  Base address of PC card. Must match jumper setting:
  ===================================================

  LK1    LK2    ADDRESS

  IN     IN     0x200 - 0x23F
  OUT    IN     0x280 - 0x3BF
  IN     OUT    0x300 - 0x33F
  OUT    OUT    0x380 - 0x3BF
*/

#define CAMAC_BASE 0x300

/*
  Interrupt request selection. Must match jumper setting:
  =======================================================

  LINK IN       IRQ line    Software interrupt

   9            IRQ 5       INT 0x0D (13)
  10            IRQ 3       INT 0x0B (11)
  11            IRQ 9       INT 0x02 (Cascaded)
  12            IRQ 12      INT 0x02 (Cascaded)
*/

#define LAM_INT 0x0D

/*------------------------------------------------------------------*/

#ifndef CAMAC_BASE_DEFINED
#define CAMAC_BASE_DEFINED
unsigned short _cam_base = CAMAC_BASE;
#else
extern unsigned short _cam_base;
#endif

/*------------------------------------------------------------------*/

/*
   macro                purpose

   output               address  data     status
   ======               =======  ======   ========

   CAMO8              : n, a, f   8 bit   -
   CAMO               : n, a, f  16 bit   -
   CAMO24             : n, a, f  24 bit   -

   CAMOQ8             : n, a, f   8 bit   X, Q
   CAMOQ              : n, a, f  16 bit   X, Q
   CAMOQ24            : n, a, f  24 bit   X, Q

   CAMO_RSTART        : n, a, f  16 bit   -     Start repeat cycle
   CAMO_REPEAT        : -        16 bit   -     Repeat cycle

   input
   =====

   CAMI8              : n, a, f   8 bit   -
   CAMI               : n, a, f  16 bit   -
   CAMI24             : n, a, f  24 bit   -

   CAMIQ8             : n, a, f   8 bit   X, Q
   CAMIQ              : n, a, f  16 bit   X, Q
   CAMIQ24            : n, a, f  24 bit   X, Q

   CAMI_RSTART        : n, a, f  16 bit   -     Start repeat cycle
   CAMI_REPEAT        : -        16 bit   -     Repeat cycle
   CAMI_LOOP          : n, a1, a2, f  16 bit    Loop subaddress

   misc
   ====

   CAMAC_N            : n                 set n for following operations
   CAMAC_A            : a                 set a for following operations
   CAMAC_F            : f                 set f for following operations
   CAMAC_S            :                   read back status in AX

   CAMAC_SET_CRATE                        set camac crate 0..3

   CAMAC_SET_INHIBIT
   CAMAC_CLEAR_INHIBIT
   CAMAC_C_CYCLE
   CAMAC_Z_CYCLE
   CAMAC_IRQ_ENABLE
   CAMAC_IRQ_DISABLE
   CAMAC_SET_IRQ_BIT(station)             set mask bit for station
   CAMAC_CLEAR_IRQ_BIT(station)           clear mask bit for station
   CAMAC_READ_LAM()                       read encoded lam

*/

/*------------------------------------------------------------------*/

#define CAMAC_N(_n)  _outp((short)(_cam_base+8), (char) _n)

#define CAMAC_A(_a)  _outp((short)(_cam_base+6), (char) _a)

#define CAMAC_F(_f)  _outp((short)(_cam_base+10), (char) _f)

#define CAMAC_OB(_d) _outp((short)(_cam_base), (char) _d)

#define CAMAC_OW(_d) \
                     _outp((short)(_cam_base), (char) _d); \
                     _outp((short)(_cam_base+2), (char) ((_d) >> 8))

#define CAMAC_OD(_d) \
                     _outp((short)(_cam_base), (char) _d); \
                     _outp((short)(_cam_base+2), (char) ((_d) >> 8)) \
                     _outp((short)(_cam_base+4), (char) ((_d) >> 16))

#define CAMAC_S      _inp((short)(_cam_base+6))

#define CAMAC_Q(_x, _q)                 \
{ char _status;                         \
  _status = _inp((short)(_cam_base+6)); \
  *_q = _status & 1;                     \
  *_x = _status >> 7; }

/*------------------------------------------------------------------*/

#define CAMO8(_n, _a, _f, _d) \
{ CAMAC_OB(_d);      \
  CAMAC_A(_a);       \
  CAMAC_N(_n);       \
  CAMAC_F(_f);       \
  CAMAC_S; }

#define CAMO(_n, _a, _f, _d) \
{ CAMAC_OW(_d);      \
  CAMAC_A(_a);       \
  CAMAC_N(_n);       \
  CAMAC_F(_f);       \
  CAMAC_S; }

#define CAMO24(_n, _a, _f, _d) \
{ CAMAC_OD(_d);      \
  CAMAC_A(_a);       \
  CAMAC_N(_n);       \
  CAMAC_F(_f);       \
  CAMAC_S; }

/*------------------------------------------------------------------*/


#define CAMOQ8(_n, _a, _f, _d, _x, _q)  \
{ char _status;                         \
  CAMO8(_n, _a, _f, _d);                \
  CAMAC_Q(_x, _q); }

#define CAMOQ(_n, _a, _f, _d, _x, _q)   \
{ CAMO(_n, _a, _f, _d);                 \
  CAMAC_Q(_x, _q); }

#define CAMOQ24(_n, _a, _f, _d, _x, _q) \
{ CAMO24(_n, _a, _f, _d);               \
  CAMAC_Q(_x, _q); }

/*------------------------------------------------------------------*/

#define CAMI8(_n, _a, _f, _d) \
{ CAMAC_A(_a);       \
  CAMAC_N(_n);       \
  CAMAC_F(_f);       \
  CAMAC_S;           \
  *(_d) = (unsigned char) _inp(_cam_base); }

#define CAMI(_n, _a, _f, _d) \
{ CAMAC_A(_a);       \
  CAMAC_N(_n);       \
  CAMAC_F(_f);       \
  CAMAC_S;           \
  *(_d) = (unsigned int) (_inp(_cam_base) + (_inp((short)(_cam_base+2)) << 8)); }

#define CAMI24(_n, _a, _f, _d) \
{ CAMAC_A(_a);       \
  CAMAC_N(_n);       \
  CAMAC_F(_f);       \
  CAMAC_S;           \
  *(_d) = (unsigned long) (_inp(_cam_base) + (_inp((short)(_cam_base+2)) << 8) + (_inp((short)(_cam_base+4))<<16));}

/*------------------------------------------------------------------*/

#define CAMIQ8(_n, _a, _f, _d, _x, _q) \
{ CAMI8(_n, _a, _f, _d)                \
  CAMAC_Q(_x, _q); }

#define CAMIQ(_n, _a, _f, _d, _x, _q) \
{ CAMI(_n, _a, _f, _d)                \
  CAMAC_Q(_x, _q); }

#define CAMIQ24(_n, _a, _f, _d, _x, _q) \
{ CAMI24(_n, _a, _f, _d)                \
  CAMAC_Q(_x, _q); }

/*------------------------------------------------------------------*/

#define CAMI_RSTART(_n, _a, _f, _d) \
{ unsigned _tmp;     \
  CAMAC_A(_a);       \
  CAMAC_N(_n);       \
  CAMAC_F(_f);       \
  CAMAC_S;           \
  *(_d) = _inp(_cam_base) + (_inp((short)(_cam_base+2)) << 8); \
  _inp((short)(_cam_base+12))

#define CAMI_REPEAT(_d)  *(_d)=_inpw((short)(_cam_base+12))

/*------------------------------------------------------------------*/

#define CAMIQ_RSTART(_n, _a, _f) \
  CAMAC_A(_a);       \
  CAMAC_N(_n);       \
  CAMAC_F(_f);       \
  CAMAC_S;           \
  _inp((short)(_cam_base+12))

#define CAMIQ_REPEAT(_q) \
  _inp((short)(_cam_base+12)); \
  *(_q) = _inp((short)(_cam_base+6)) & 1;

#define CAMI_WAITQ(_n, _a, _f, _t) \
  {                  \
  WORD x, q;         \
                     \
  CAMAC_N(_n);       \
  CAMAC_A(_a);       \
  CAMAC_F(_f);       \
  CAMAC_Q(&x, &q);   \
                     \
  while (!q && *_t)  \
    {                \
    CAMAC_F(_f);     \
    CAMAC_Q(&x, &q); \
    (*_t)--;         \
    }                \
  }

/*------------------------------------------------------------------*/

#define CAMO_RSTART(_n, _a, _f, _d) CAMO(_n, _a, _f, _d)

#define CAMO_REPEAT(_d) _outpw((short)(_cam_base+12), _d)

/*------------------------------------------------------------------*/

#define CAMI_LOOP(_n, _a1, _a2, _f, _d) \
  {                                     \
  WORD a, *pdata;                       \
                                        \
  pdata = (_d);                         \
  CAMI(_n, _a1, _f, pdata++);           \
                                        \
  for (a=_a1+1 ; a<=_a2 ; a++)          \
    {                                   \
    CAMAC_A(a);                         \
    CAMAC_F(_f);                        \
    CAMAC_S;                            \
    *(pdata++) = (int) _inp(_cam_base) + (_inp((short)(_cam_base+2)) << 8); \
    }                                   \
  }

/*------------------------------------------------------------------*/

#define CAMAC_INIT() \
  ss_directio_init(); \
  ss_directio_give_port(_cam_base, _cam_base+0x3F); \
  ss_directio_exit()

#define CAMAC_SET_CRATE(_crate) \
  _cam_base = CAMAC_BASE + 0x10*_crate; \

#define CAMAC_SET_INHIBIT()   _outp((short)(_cam_base+10), 33);

#define CAMAC_CLEAR_INHIBIT() _outp((short)(_cam_base+10), 32);

#define CAMAC_C_CYCLE()       _outp((short)(_cam_base+10), 36);

#define CAMAC_Z_CYCLE()       _outp((short)(_cam_base+10), 34);

#define CAMAC_IRQ_ENABLE()    _outp((short)(_cam_base+10), 41);

#define CAMAC_IRQ_DISABLE()   _outp((short)(_cam_base+10), 40);

#define CAMAC_SET_IRQ_BIT(station) _outp((short)(_cam_base+10), 64+station);

#define CAMAC_CLEAR_IRQ_BIT(station) _outp((short)(_cam_base+10), 128+station);

#define CAMAC_READ_LAM(_l)    _l = _inp((short)(_cam_base+8));
